from distutils.core import setup
setup(
  name = 'npeb',
  packages = ['npeb'],
  version = '0.0.2', 
  license='MIT',   
  description = 'Nonparametric Empirical Bayes',
  author = 'Jake A. Soloff',
  author_email = 'jakesoloff@gmail.com',     
  url = 'https://github.com/jake-soloff/NPEB', 
  download_url = 'https://github.com/jake-soloff/NPEB/archive/refs/tags/v_001.tar.gz',
  keywords = ['Empirical Bayes', 'NPMLE', 'Mixture models'], 
  install_requires=[      
          'numpy',
          'scipy',
          'scikit-learn',
          'pandas',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      
    'Intended Audience :: Science/Research',    
    'Intended Audience :: Developers',    
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',  
    'Programming Language :: Python :: 3', 
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
 ],
)
